#!/usr/bin/perl
#
#          Rogue Detect version 1.0
# DHCP Server Detection Software
#
# Rogue Detect is a quick&dirty DHCP Server Detector, designed
# for use on networks such as ResNet's where a rogue DHCP server
# can cause massive damage and helpdesk phonecalls.
#
# It uses the included DHCP Server Detector library DHCPDetect.pm
# 
# Its tested with Debian Linux (Woody +some 'unstable' packages)
#
# For info see http://oss.oregonstate.edu/roguedetect
#
# This program is made possible by the Oregon State University
# Open Source movement.
#
# Oregon State University
# Open Minds. Open Doors. Open Source.
# http://oregonstate.edu
#
# COPYRIGHT
#
# Written by:
#  Alex Schumann, Oregon State University, alex.schumann@oregonstate.edu
#  initial release: Sept 10, 2003
#
# LISCENSE
#
# This program is free software; you can redistribute it and/or modify it under the terms of the 
# GNU General Public License as published by the Free Software Foundation; either version 1, or 
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See either the 
# GNU General Public License or the Artistic License for more details.
#
# This is an attempt to detect dhcp servers on
# the resnet, and take appropreate action.

# Object methods to create a DHCP packet
#use Net::DHCP::Packet;
use Getopt::Long;
use POSIX;
use Net::SMTP;

use strict;

# Set this path to where the DHCPDetect.pm file will
# be on your system, if you want to execute it from
# outside the directory its in. Eg
use lib qw(/data/scripts/roguedetect/DHCPDetect);
# use lib qw(DHCPDetect);
use DHCPDetect;
use OUI;

my $version = "1.0";


#########################################
# CONFIG SECTION
# 
# list of MAC addresses of authorized DHCP servers on your network
# put the hwa of any 'legitimate' dhcp servers or dhcp forwarding 
# routers in this list. Anything you want to ignore.
my @approvedservers = ( # THESE _MUST_ be lowercase..
                        '00:07:e9:17:88:93', # some host
                        '00:07:e9:17:88:96', # some other host
                        '00:07:e9:17:87:6f', # this guy has a legit dhcp server
                        );

# Interfaces - put your HWAs here for each interface
# on your system.  The bait is sent from these addresses.
# FIXME - cant we just use our real one(s) automagically???
my %ifaces = ('eth0' => '00:B0:01:23:45:67',
              'eth1' => '00:B0:11:23:45:67',
              'eth2' => '00:07:E9:17:88:96',
              'eth3' => '00:07:E9:17:87:6F',
              'eth4' => '00:07:E9:17:8D:4F');
# Comment this out this next line after you fix the hwas above and
# genreally customized the script
#die("You must edit this script before you can run it! Comment this out after you have done so.\n");

# The default action is to email this address. Change it to yours.
# take care to use '' not "" or escape @.
my $alert_email = 'root@localhost';

#########################################
# COMMANDLINE VARIABLE DEFAULTS 
# 
# You don't need to edit these, just use the commandline arguments
# to override them
#
# in continuous mode, how often 
# to send DHCPDISCOVER packets (minutes)
my $frequency = 10;
# Debug level. 0=errors only, 1=info, 2=debug, 5 = highest
#              If debug is set higher than 0, no email gets sent.
my $debug = 0;
# In continuous mode, should the program fork into a background process?
my $daemon = '1';
# Give info about whats going on
my $verbose = 0;
# Which interface to use
my $interface = 'eth0';
# Continuous runs a non-stop scan. Non-Continuous mode
# scans for <waitfor> seconds and quits (eg, to run once, or from a cron script)
my $continuous = '0';
# How long to wait after a DHCPDISCOVER (noncontinuous operation only)
my $waitfor = '45';  # Long enough for MACfinder.pl to finish...
# Save hwaddrs in a temp dir to avoid sending multiple emails for the same server?
my $save = 0;

# Trailing slash is a must..
my $hwadir = "/tmp/dhcp_hwaddrs/"; #created automatiaclly if not exists

my $givehelp = 0;

## End of command line defaults
###################################

sub approvedserver {
    my $self = shift;

    my $s_hwa = $self->svr_hwa();
    my $ok = 0;

    foreach $a (@{$self->approved}) {
        if($a eq $s_hwa) {
            if($debug > 1) {
                print "DEBUG: $s_hwa is an approved server, ignoring..\n";
            }
            return(1);
        }
    }
    if($debug > 1) {
        print "DEBUG: $s_hwa is not on the approved server list!\n";
    }
    return(0);
}

my $result = GetOptions("debug=i" => \$debug, 
                        "freq=i"  => \$frequency,
                        "verbose" => \$verbose,
                        "iface=s"   => \$interface,
                        "continuous!" => \$continuous,
                        "waitfor=i" => \$waitfor,
                        "daemon!" => \$daemon,
                        "save!"  => \$save,
                        "help" => \$givehelp);



if($givehelp)
{
   print "
Rogue Detect version $version
DHCP Server Detector

SYNTAX:
  dhcpdetect.pl --continuous [--freq=x] [--iface=interface] \\
               [--nodaemon] [--verbose] [--debug=x]
OR
  dhcpdetect.pl [--waitfor=x] [--iface=interface] [--verbose] [--debug=x]


  --continuous          Cause the program to run in a loop, sending 
                           DHCPDISCOVER occationally and listning for replies
  --nocontinuous        (default) Send only one DHCPDISCOVER and wait only
                           long enough for a response. Then quit. (for 
                           commandline or crontab use)
  --freq=x              In continuous mode only: how long between 
                           DHCPDISCOVER packets in seconds
  --waitfor=x           In normal (non-continuous) mode only: how long to
                           wait after sending a DHCPDISCOVER for replies.
  --iface=interface     Network interface to use (eg eth0). Unfortunately
                           only one can be done at a time (sorry) run 
                           multiple copies for now.
  --daemon/--nodaemon   In continuous mode only: weather or not the program
                           should fork and go into the background. Default 
                           is yes.
  --save/--nosave       Save rogue hw addresses in a temp dir to avoid sending
                        multiple emails about the same address.
  --verbose             Give details -- currently not well supported. use 
                           debug=1 instead.
  --debug=x             Debug level. 1-5 supported. 1 is good for info. higher
                           levels give more information.
  \n\n";
  exit;
}

#make sure the hwawdir path exists
if ( $save && ! -d $hwadir )
{
   mkdir($hwadir) || die "Cannot mkdir $hwadir: $!";
}

if($debug)
{
   $verbose = 1;
   $daemon = 0;
}

if(!$continuous)
{
   $daemon = 0;
}
if($daemon)
{
   $verbose = 0;
}

if($debug > 0)
{
  print "Options: Debug:$debug frequency:$frequency verbose:$verbose iface:$interface continuous:$continuous waitfor:$waitfor daemon:$daemon help$givehelp\n";
}

if($verbose)
{
   print "DHCP Server Detector version $version starting.\n";
}


# Fork us into the background as daemon, if requested
if($daemon)
{
    defined(my $pid = fork) or die "Cannot fork: $!";
    if($pid) {
        exit;  # Parent quits..
    }
    close STDOUT; close STDERR; close STDIN;
    setsid() or die "Cannot start a new session: $!\n";
}

my $macaddr = $ifaces{$interface};
# Open a new DHCPDetect object..
my $dhcp = new DHCPDetect( macaddr => $macaddr,
                           interface => $interface, 
                           debug => $debug,
                           timeout => 2,
                           approved => \@approvedservers);

#Spawn a child to listen for dhcp responses..
if($debug)
{
   print "Spawning a dhcp packet listner..\n";
}
defined(my $pid = fork) or die "Cannot fork: $!";
unless ($pid) {
    #child (listener)

    # Read packets in a loop until killed
    while(1)
    {
        # get 1 packet from scanner
        $dhcp->scan();
        if($debug >1)
        {
           print "scan() returned.\n";
        }
        my $server = $dhcp->server_ip;
        my $reply = $dhcp->reply;
        my $hwa = $dhcp->svr_hwa;
        if($dhcp->server_ip eq '') {
            if($debug > 0) {
                print "Packet recieved but has no source ip\n";
            }
        }
        elsif($dhcp->reply ne 'offer') {
            if($debug > 0) { 
               print "Packet from " . $dhcp->svr_hwa . " is a '$reply' not an 'offer'\n";
            }
        }
        elsif(approvedserver($dhcp)) {
            if($debug > 0) {
                print "Packet from ". $dhcp->svr_hwa ." is identified as from an official DHCP server\n";
            }
        }
        else { # all above conditions failed, so its a rogue DHCP Server!
            do_alert($hwa, $server, $reply);
        }
    }
    die("Child is done\n");
}

#parent continues on to do periodic sends
if($debug > 1)
{
    print "Parant says: CHILD PID IS $pid\n";
}


# Catch signals so we can kill the child first
$SIG{'QUIT'} = \&sighandler;
$SIG{'INT'} = \&sighandler;
$SIG{'HUP'} = \&sighandler;
$SIG{'TERM'} = \&sighandler;

# Fire up our DHCPDISCOVER loop
sleep(1); # give the listener a chance to warm up
while(1)
{
    if($debug)
    {
       print "Sending a DHCPDISCOVER packet..\n";
    }
    $dhcp->bait();

    # If were not in continuous mode,
    # wait waitfor seconds and 
    # quit.
    if(!$continuous)
    {
        sleep($waitfor);
        if($debug)
        {
           print "Scan completed. Exiting\n";
        }
        kill('TERM',$pid);
        exit;
    }
    else
    {
        #
        # If we are in continuous mode, repeatedly scan
        #
        # Wait frequency minutes, while checking 
        # child process for having died.
        my $i = 0;
        for($i = 0; $i < $frequency * 3600; $i++)
        {
        
            my $res = waitpid($pid, &POSIX::WNOHANG);
            if($res != 0)
            {
#                print "waitpid returned $res, Child no longer exists!\n";
                die("Child nolonger exists\n");
            }
            sleep(1); 
        }
    }
}

sub send_mail
{
    my $to = shift;
    my $subject = shift;
    my $body = shift;

    my $smtp = Net::SMTP->new('localhost', 
                              Timeout => 20,
                              Debug => $debug);

    if($debug)
    {
        print "Sending alert email:\n";
        print "To: $to\n";
        print "Subject: $subject\n";
        print "\n";
        print "$body\n\n";
    }
    $smtp->mail('dhcpserverdetector@blank.uhds.oregonstate.edu');
    $smtp->to($to);

    $smtp->data();
    $smtp->datasend("To: $to\n");
    $smtp->datasend("Subject: $subject\n");
    $smtp->datasend("\n");
    $smtp->datasend("$body\n");
    $smtp->dataend();
    $smtp->quit;
}

#make sure the hwawdir path exists
if ( $save &&  ! -d $hwadir )
{
   mkdir($hwadir) || die "Cannot mkdir $hwadir: $!";
}

sub do_alert 
{
    my $hwa = shift; 
    my $server = shift;
    my $reply = shift;

    my $timestamp = localtime();

    if($debug) {
       print "Detected a rogue DHCP server! HWA: $hwa\n";
    }

    if($save && -e $hwadir . $hwa) {
        if($debug) {
            print "Already send email out for this rogue DHCP server.\n";
        }
        return;
    }
    if($save){
        open(FILEHANDLE, ">" . $hwadir . $hwa) || die('cannot open file: '.$!);
        close(FILEHANDLE);
    }
    my $body = "Rogue DHCP Server Detected on $interface\n";
    $body .= "HWA: $hwa (" . OUI::oui($hwa). ") - IP: $server\n";
    #########################################
    ## Add your own additional actions here!
    ##   $hwa is the hardware address of the dhcp server
    ##   $server is the IP that server claims to be
    ##   $interface is the ethernet device it was seen on (--iface)
    ##   anything you add to $body will be added to the email message
    ## Ideas:
    ##   * Execute a script to search your switches for what port that HWA is on
    ##   * Log to a database
    ##   * auto-disable the port
    ##   * (shown) add timestamp to email body
    ##   * ??
    ## Custom user actions:

    $body .= "$timestamp\n";
    
    ## End of custom user actions
    #########################################
    my $subject = "Rogue DHCP Server Detected on $interface";
    if(!$debug)
    {
       send_mail($alert_email, $subject, $body);
    }
}


sub sighandler
{
    my $sig = shift;
    if($debug)
    {
       print "GOT SIG$sig. Shutting down CHILD ($pid) and self\n";
    }
    kill('TERM', $pid);
    die("Deth by signal\n");
}


